import cattura.agents.configuration.model.vos.plugin.PluginInfo
import cattura.agents.configuration.model.interfaces.Plugin.Inject
import cattura.agents.configuration.model.interfaces.Plugin.On
import org.puremvc.java.multicore.patterns.facade.Facade
import cattura.agents.configuration.model.proxies.PluginProxy
import cattura.agents.configuration.model.interfaces.Plugin
import cattura.agents.configuration.model.proxies.EnvironmentProxy
import cattura.agents.configuration.model.proxies.capture.MediaProcessingProxy
import com.cattura.agents.bridge.model.GraphBuilder
import groovy.io.FileType
import java.util.zip.ZipFile
import java.util.zip.ZipOutputStream
import java.util.zip.ZipEntry
import cattura.utils.parsing.JSON
import cattura.agents.configuration.model.vos.watchers.DiskSpaceWatcher
import cattura.agents.configuration.model.vos.watchers.NetworkStatusWatcher
import cattura.agents.configuration.model.proxies.UpdaterProxy
import cattura.agents.configuration.model.proxies.UpdaterProxy.VersionInfo
import cattura.agents.configuration.model.proxies.SettingsProxy
import cattura.agents.configuration.model.properties.SettingNames
import java.text.SimpleDateFormat
import java.text.ParseException
import java.util.Arrays
import groovy.json.JsonOutput
import com.google.common.io.Files
import com.google.common.base.Charsets
import cattura.utils.io.ProcessHelper
import cattura.agents.configuration.model.utilities.ProcessUtilities
import com.google.common.collect.TreeTraverser
import com.google.common.collect.Maps
import com.cattura.share.processor.Processor.Subscribe
import com.cattura.share.processor.Processor.Factory.Service
import cattura.agents.configuration.model.vos.events.capture.CaptureCrashEvent
import com.cattura.agents.bridge.model.MediaProcessingModule.Session
import jodd.io.ZipUtil
import com.cattura.share.processor.graph.Graph
import java.util.Collections
import com.cattura.agents.bridge.model.MediaProcessingModule
import cattura.agents.configuration.model.vos.events.Event
import cattura.utils.os.OSSystemState
import com.google.common.cache.Cache
import com.google.common.cache.CacheBuilder
import cattura.agents.configuration.model.vos.IncidentReport
import java.util.concurrent.atomic.AtomicBoolean
import cattura.utils.time.Time.VO.Timeout
import java.util.UUID
import cattura.agents.configuration.model.vos.events.Event
import cattura.agents.configuration.controller.facade.ApplicationFacade
import java.util.concurrent.TimeUnit

class LogDownloaderPlugin implements Plugin
{
    public static @Inject.Info PluginInfo _info
    public static @Inject.Facade Facade _facade
    private static final String URI_PREFIX = "/logs/"

    private static final SimpleDateFormat LOG_FILE_TIME_FORMATTER = new SimpleDateFormat("y-M-d_h-m-s-a")
    private static final SimpleDateFormat DATE_LOG_FORMATTER = new SimpleDateFormat("MM/dd/yyyy h:m:s a")
    
    private ProcessHelper _logReader;
    private ProcessHelper _networkDebugger;

    private File _logDir
    private File _osLogDir
    private File _captureSessionDir
    private File _storageDir
    private File _crashDir

    private JSON _json = new JSON.Default()

    /*
     * Exception tracking
     */
    final Map<Date, Throwable> _exceptions = [:]

    /*
     * Log archives created through createLogArchiveAsync
     */
    final Map<String, File> _archives = [:]

    /*
     * Incident log archive generation
     */
    final AtomicBoolean _currentlyDumpingLogs = new AtomicBoolean(false);
    final Cache<IncidentReport, File> _incidentLogArchives = CacheBuilder.newBuilder()
        .expireAfterAccess(1, TimeUnit.MINUTES)
        .removalListener({
            _info.getLogger().info("Cleaning up stray log file: ${it.value}")
            it.value.delete()
        })
        .build();
    
    @On.Load
    public void onLoad ()
    {
        final EnvironmentProxy environment = _facade.retrieveProxy(EnvironmentProxy.NAME)
        _logDir = environment.getLogPath()
        _captureSessionDir = environment.getCaptureSessionPath()
        _storageDir = environment.getStorageDirPath()
        _crashDir = new File(environment.getRootPath(), "crashes")
        _osLogDir = new File("/var/log")
        _info.getLogger().info("Log Downloader Plugin started")
        _logReader = ProcessUtilities.createPrivilegedProcess(new ProcessHelper(_info.getResource("scripts", "read_log").getAbsolutePath()));
        _networkDebugger = ProcessUtilities.createPrivilegedProcess(new ProcessHelper(_info.getResource("scripts", "debug_network").getAbsolutePath()));
    }

    /*
     * Set up HTTP Handler when the web server is constructed
     */
    @Subscribe
    public void onEvent (final Event $event)
    {
        if ($event.getClass().getSimpleName() == "WebServerEvent")
        {
            def LogDownloaderHttpHandler = this.class.classLoader.parseClass(_info.getResource("LogDownloaderHttpHandler.groovy"))
            $event.getWebServer().add(LogDownloaderHttpHandler.newInstance(this, _storageDir))
        }
    }

    /*
     * Exception monitor
     */
    /*@Subscribe(type=IncidentEvent.class)
    public void onException (final Throwable $throwable)
    {
        _exceptions.put(new Date(), $throwable)
    }*/

    @Subscribe(type=CaptureCrashEvent.class)
    public void onCaptureCrashEvent (final CaptureCrashEvent $event)
    {
        final Session session = $event.getSession()
        final File logsDir = getExtraLogsDirectoryForSession(session.getRoot())
        _info.getLogger().info("Log Downloader Plugin is collecting logs for failed capture in [${logsDir}]")

        final File zip = createLogArchive([
            from: session.getStartTime()
        ])

        logsDir.mkdirs()
        ZipUtil.unzip(zip, logsDir)
        zip.delete()
    }

    /*
     * Creates a log dump for an incident report.
     */
    @Service
    public File createLogArchiveForIncident (final @Service.Parameter(name="incident") IncidentReport $incident)
    {
        return _incidentLogArchives.get($incident, {
            if (!_currentlyDumpingLogs.compareAndSet(false, true))
            {
                throw new RuntimeException("Was unable to generate info dump for incident because log downloader plugin is currently dumping logs.")
            }
            
            try
            {
                String description = "" + $incident.getMessage();
                
                if ($incident.getException() != null)
                {
                    description += " " + $incident.getException().getLocalizedMessage() + " " + $incident.getException().getClass().getName();
                }
                
                description = description.toLowerCase();
                
                return createLogArchive([
                    includeExceptions: false,
                    useQuickLsof: true,
                    includeNetworkDebug: description.contains("java.net") || description.contains("network") || description.contains("connection")
                ])
            }
            finally
            {
                _currentlyDumpingLogs.set(false)
            }
        });
    }
    
    @Service
    public String createLogArchiveAsync (final @Service.Parameter(name="params") Map $params)
    {
        final String key = UUID.randomUUID().toString() + ".zip"
        Timeout.countdown(0, {
            final File archive = createLogArchive($params)
            _archives.put(key, archive)
            _facade.sendNotification(ApplicationFacade.EMIT_EVENT, new Event(
                "LogDownloaderLogsReadyEvent",
                [ key: key ]
            ))
        })
        return key
    }
    
    /*
     * Collects logs timestamped from $from to $to and returns a zip file
     * containing all of them.
     */
    @Service
    public File createLogArchive (final @Service.Parameter(name="params") Map $params)
    {
        /*
         * Log file parameters
         */
        final Date to = argumentAsDate($params.get("to"))     ?: new Date() // default: now
        final Date from = argumentAsDate($params.get("from")) ?: new Date(to.getTime() - 3600000) // default: one hour from "to"
        final boolean includeCatturaLogs = $params.containsKey("includeCatturaLogs") ? $params.get("includeCatturaLogs") : true
        final boolean includeCrashLogs = $params.containsKey("includeCrashLogs") ? $params.get("includeCrashLogs") : true
        final boolean includeOsLogs = $params.containsKey("includeOsLogs") ? $params.get("includeOsLogs") : true
        final boolean includeExceptions = $params.containsKey("includeExceptions") ? $params.get("includeExceptions") : true
        final boolean includeSystemInformation = $params.containsKey("includeSystemInformation") ? $params.get("includeSystemInformation") : true
        final boolean includeSystemStateInformation = $params.containsKey("includeSystemStateInformation") ? $params.get("includeSystemStateInformation") : true
        final boolean includeThreadDumps = $params.containsKey("includeThreadDumps") ? $params.get("includeThreadDumps") : true
        final boolean includeNetworkDebug = $params.containsKey("includeNetworkDebug") ? $params.get("includeNetworkDebug") : true

        /*
         * Get all (system and capture session) logs between dates
         */
        final Collection<FoundLogFile> catturaLogs = includeCatturaLogs ? getCatturaSystemLogsBetween(from, to) : Collections.EMPTY_LIST
        final Collection<FoundLogFile> crashLogs = includeCrashLogs ? getLogsFromDirBetween(_crashDir, from, to) : Collections.EMPTY_LIST
        final Collection<FoundLogFile> osLogs = includeOsLogs ? getLogsFromDirBetween(_osLogDir, from, to, false).findAll({
            // Limit to log files that we might care about
            return it.logFile.name == "syslog"   || it.logFile.name.startsWith("syslog.") ||
                   it.logFile.name == "kern.log" || it.logFile.name.startsWith("kern.log.")
        }) : Collections.EMPTY_LIST

        /*
         * Create zip file
         */
        return createZip() { $zipOut ->
            /*
             * Insert system logs
             */
            addFilesToZip($zipOut, catturaLogs, "logs/")

            /*
             * Insert crash logs
             */
            addFilesToZip($zipOut, crashLogs, "crashes/")

            /*
             * Insert OS logs
             */
            addFilesToZip($zipOut, osLogs, "osLogs/")

            /*
             * Insert installed plugins
             */
            addPluginInformationToZip($zipOut)

            /*
             * Insert exceptions
             */
            if (includeExceptions)
            {
                addExceptionsToZip($zipOut) { $date, $exception ->
                    return dateIsBetween(from, $date, to)
                }
            }

            /*
             * Insert system information
             */
            if (includeSystemInformation)
            {
                addSystemInformationToZip($zipOut)
            }

            /*
             * Insert system state
             */
            if (includeSystemStateInformation)
            {
                addSystemStateInfoToZip($zipOut, $params)
            }

            /*
             * Insert thread dumps
             */
            if (includeThreadDumps)
            {
                addThreadDumpsToZip($zipOut)
            }

            /*
             * Insert network debug log
             */
            if (includeNetworkDebug)
            {
                addProcessOutputToZip($zipOut, _networkDebugger.clone(), "network_debug.log")
            }
        }
    }

    /*
     * Zips up all logs for the capture session located at $root.
     */
    public File createCaptureSessionLogArchive (final File $root)
    {
        /*
         * Directory where the "extra logs" are stored in case of a failure
         */
        final File extraLogsDir = getExtraLogsDirectoryForSession($root)

        /*
         * Get logs for this capture session
         */
        final Collection<CaptureSessionLog> captureSessionLogs = getCaptureSessionLogsInDir($root)

        /*
         * Create zip file
         */
        return createZip() {
            addCaptureSessionLogsToZip(it, captureSessionLogs, false)

            /*
             * Add graph to zip file.
             */
            it.putNextEntry(new ZipEntry("graph.json"))
            it << serializeGraphAtRootForZip($root)
            it.closeEntry()

            if (extraLogsDir.isDirectory())
            {
                addFilesToZip(it, getLogsFromDirBetween(extraLogsDir, new Date(0), new Date()), "extraLogs/")
            }
        }
    }

    public File getAsyncCreatedLogArchive (final String $key)
    {
        return _archives[$key]
    }
    
    /*
     * Creates a zip file. The ZipOutputStream is passed to the $closure,
     * which can add entries to the zip.
     *
     * When the method returns, the stream is closed.
     */
    private File createZip (final Closure $closure)
    {
        final File zipFile = File.createTempFile("catturalogs",".zip")
        final ZipOutputStream zipOut = new ZipOutputStream(new FileOutputStream(zipFile))
        $closure(zipOut)
        zipOut.close()
        return zipFile
    }

    private void addFilesToZip (final ZipOutputStream $zipOut, final Collection<FoundLogFile> $files, final String $prefix = null)
    {
        for (final FoundLogFile file : $files)
        {
            final String prefix = ($prefix ?: "")
            if (!file.getLogFile().isFile())
            {
                continue
            }

            addFileToZip($zipOut, file.getLogFile(), prefix + file.getEntryName())
        }
    }

    private File getExtraLogsDirectoryForSession (final File $sessionDir)
    {
        return new File("extraLogs", $sessionDir)
    }

    /*
     * Adds the given $logs to the ZipOutputStream $zipOut.
     *
     * If $usePrefix is true, each entry is prefixed with the name of the capture session root directory.
     *     Use this if you are adding logs from multiple capture sessions to the zip.
     */
    private void addCaptureSessionLogsToZip (final ZipOutputStream $zipOut, final Collection<CaptureSessionLog> $logs, final boolean $usePrefix = true)
    {
        for (final CaptureSessionLog log : $logs)
        {
            final String prefix = $usePrefix ? "${log.rootFile.name}/" : ""
            $zipOut.putNextEntry(new ZipEntry("${prefix}${log.zipEntryName}"))
            try
            {
                log.logFile.withInputStream() {
                    $zipOut << it
                }
            }
            catch (final Exception $exception)
            {
                _info.getLogger().info("Log Downloader Plugin failed to read `${log.getLogFile()}`: ${$exception.message}")
            }
            $zipOut.closeEntry()
        }
    }

    /*
     * Adds logged exceptions to the zip file as a text file (for humans) and JSON file (for programmatic consumption).
     */
    private void addExceptionsToZip (final ZipOutputStream $zipOut, final Closure $closure = { return true }) {addExceptionsToZip($zipOut, "", $closure)}
    private void addExceptionsToZip (final ZipOutputStream $zipOut, final String $prefix, final Closure $closure = { return true })
    {
        def jsonObject = []
        def textObject = new StringWriter()
        def textWriter = new PrintWriter(textObject)

        /*
         * Loop through exceptions and construct output
         */
        _exceptions.each() { date, exception ->
            if (!$closure(date, exception))
            {
                return
            }

            jsonObject.add(exceptionToMap(date, exception))
            textWriter << "${date} "
            exception.printStackTrace(textWriter)
            textWriter << "---\r\n"
        }

        /*
         * Write outputs to zip file
         */
        $zipOut.putNextEntry(new ZipEntry("${$prefix}exceptions.json"))
        $zipOut << JsonOutput.prettyPrint(_json.stringify(jsonObject).getOrRethrow())
        $zipOut.closeEntry()

        $zipOut.putNextEntry(new ZipEntry("${$prefix}exceptions.txt"))
        $zipOut << textObject
        $zipOut.closeEntry()
    }

    /*
     * Adds system information to the zip file. This includes network
     * status, unit name, build number, disk space, etc.
     */
    private void addSystemInformationToZip (final ZipOutputStream $zipOut, final String $entryName = "systeminfo.json")
    {
        final EnvironmentProxy environment = _facade.retrieveProxy(EnvironmentProxy.NAME)
        final UpdaterProxy updater = _facade.retrieveProxy(UpdaterProxy.NAME)
        final SettingsProxy settings = _facade.retrieveProxy(SettingsProxy.NAME)
        final DiskSpaceWatcher diskWatcher = environment.getDiskSpaceWatcher()
        final NetworkStatusWatcher networkWatcher = environment.getNetworkStatusWatcher()
        final VersionInfo version = updater.getVersionInfo()
        def jsonObject = [:]

        /*
         * Unit information
         */
        jsonObject.unitName = settings.getUserSettings().getString(SettingNames.UNIT_NAME).orElse("Unnamed Unit")

        /*
         * Disk space information
         */
        jsonObject.disk = [
            total: diskWatcher.getTotalSpace(),
            free: diskWatcher.getFreeSpace(),
            used: diskWatcher.getUsedSpace(),
            usable: diskWatcher.getUsableSpace()
        ]

        /*
         * Network information
         */
        jsonObject.network = [
            localAddress: networkWatcher.getLocalIP(),
            internetAddress: networkWatcher.getGlobalIP(),
            gateway: networkWatcher.getGateway(),
            hasInternet: networkWatcher.hasInternet(),
            subnet: networkWatcher.getSubnet(),
            networkType: networkWatcher.getNetworkType(),
            macAddress: networkWatcher.getMACAddress()
        ]

        /*
         * Environment variables
         */
        jsonObject.environment = System.getenv()

        /*
         * OS information
         */
        jsonObject.os = [
            kernel: new ProcessHelper("uname").arg("-r").call(),
            distribution: new ProcessHelper("lsb_release").arg("-s").arg("-d").call(),
            architecture: new ProcessHelper("uname").arg("-m").call(),
            machineName: new ProcessHelper("uname").arg("-n").call()
        ]

        /*
         * Software information
         */
        jsonObject.software = [
            build: version.getBuildNumber(),
            version: version.getCurrentVersion()
        ]

        /*
         * Runtime informaton
         */
        jsonObject.runtime = [
            java: [
                version: System.getProperty("java.version"),
                vendor: System.getProperty("java.vendor")
            ]
        ]

        /*
         * Write to zip file
         */
        $zipOut.putNextEntry(new ZipEntry($entryName))
        $zipOut << JsonOutput.prettyPrint(_json.stringify(jsonObject).getOrRethrow())
        $zipOut.closeEntry()
    }

    /*
     * Adds plugin information to zip.
     */
    private void addPluginInformationToZip (final ZipOutputStream $zipOut, final String $entryName = "plugins.json")
    {
        /*
         * MPA Module Plugins
         */
        final MediaProcessingModule module = _facade.retrieveProxy(MediaProcessingProxy.NAME).getCaptureModule()
        final List mpaModulePlugins = module && module.hasFoundPlugins() ? module.getPlugins() : Collections.EMPTY_LIST

        /*
         * CM Plugins
         */
        final Collection configurationModulePlugins = _facade.retrieveProxy(PluginProxy.NAME).getPluginInfos()

        /*
         * Write to zip file
         */
        $zipOut.putNextEntry(new ZipEntry($entryName))
        $zipOut << JsonOutput.prettyPrint(_json.stringify([
            processor: mpaModulePlugins.collect {
                return [
                    pluginId: it.ID,
                    title: it.title,
                    type: "processor",
                    subtype: it.type,
                    description: it.description,
                    version: it.version,
                    url: it.url
                ]
            },

            configurator: configurationModulePlugins.collect {
                return [
                    pluginId: it.pluginID,
                    title: it.title,
                    type: "configurator",
                    description: it.description,
                    version: it.version,
                    url: it.url
                ]
            }
        ]).getOrRethrow())
        $zipOut.closeEntry()
    }

    private void addSystemStateInfoToZip (final ZipOutputStream $zipOut, final Map $params)
    {
        final boolean useQuickLsof = $params.containsKey("useQuickLsof") ? $params.get("useQuickLsof") : false
        
        /*
         * Insert meminfo
         */
        addFileToZip($zipOut, new File("/proc/meminfo"), "meminfo.txt")

        /*
         * Insert cpuinfo
         */
        addFileToZip($zipOut, new File("/proc/cpuinfo"), "cpuinfo.txt")

        addProcessOutputToZip($zipOut, new ProcessHelper("top").env("COLUMNS","150").arg("-cbn1").arg("-H"), "top.txt")
        addProcessOutputToZip($zipOut, new ProcessHelper("sensors"), "sensors.txt")
        addProcessOutputToZip($zipOut, new ProcessHelper("lsmod"), "lsmod.txt")
        addProcessOutputToZip($zipOut, new ProcessHelper("ps").arg("-T").arg("auxm"), "ps.txt")

        /*
         * Insert kernel module info
         */
        addKernelModuleInformationToZip($zipOut)

        addProcessOutputToZip($zipOut, new ProcessHelper("lspci").arg("-vvv"), "lspci.txt")
        addProcessOutputToZip($zipOut, new ProcessHelper("lsusb").arg("-v"), "lsusb.txt")
        addProcessOutputToZip($zipOut, new ProcessHelper("df").arg("-h"), "df.txt")
        
        ProcessHelper lsof = new ProcessHelper("lsof")
        if (useQuickLsof)
        {
            lsof.arg("-u").arg(System.getenv("USER"))
        }
        addProcessOutputToZip($zipOut, lsof, "lsof.txt")
        
        addProcessOutputToZip($zipOut, ProcessUtilities.createPrivilegedProcess(new ProcessHelper("dmidecode")), "dmidecode.txt")
        addProcessOutputToZip($zipOut, new ProcessHelper("netstat"), "netstat.txt")
        addProcessOutputToZip($zipOut, new ProcessHelper("iostat"), "iostat.txt")
        addProcessOutputToZip($zipOut, new ProcessHelper("mpstat"), "mpstat.txt")
        addProcessOutputToZip($zipOut, new ProcessHelper("pidstat"), "pidstat.txt")
        addProcessOutputToZip($zipOut, new ProcessHelper("vmstat"), "vmstat.txt")
    }

    /*
     * Adds kernel module information for each kernel module to the zip.
     */
    private void addKernelModuleInformationToZip (final ZipOutputStream $zipOut, final String $entryName = "modinfo.json")
    {
        $zipOut.putNextEntry(new ZipEntry($entryName))
        try
        {
            $zipOut << JsonOutput.prettyPrint(_json.stringify(new ProcessHelper("lsmod").call().split("\n").collect {
                it.split(" ")[0]
            }.collectEntries {
                [it, new ProcessHelper("modinfo").arg(it).call().split("\n").collectEntries {
                    it.split(":", 2).collect { it.trim() }
                }]
            }).getOrRethrow())
        }
        catch (final Exception $exception)
        {
            _info.getLogger().info("Log Downloader Plugin failed to capture kernel module information to `${$entryName}`: ${$exception.message}")
        }
        $zipOut.closeEntry()
    }

    private void addThreadDumpsToZip (final ZipOutputStream $zipOut)
    {
        /*
         * Insert thread dump of configuration module
         */
        def cmPid = OSSystemState.getPID()
        addProcessOutputToZip($zipOut, new ProcessHelper("jstack").arg(cmPid), "threads/cm-${cmPid}.txt")

        /*
         * Insert thread dumps of MPA Modules
         */
        _facade.retrieveProxy(MediaProcessingProxy.NAME).getAllMediaProcessingModules().each {
            addProcessOutputToZip($zipOut, new ProcessHelper("jstack").arg(String.valueOf(it.getPID())), "threads/mpam-${it.getPID()}.txt")
        }
    }

    private void addFileToZip (final ZipOutputStream $zipOut, final File $input, String $name)
    {
        if ($name == null)
        {
            $name = $input.getName()
        }

        if ($input.exists())
        {
            $zipOut.putNextEntry(new ZipEntry($name))
            try
            {
                $input.withInputStream() {
                    $zipOut << it
                }
            }
            catch (final IOException $exception)
            {
                final ProcessHelper process = _logReader.clone($input.getAbsolutePath())
                addProcessOutputToZipEntry($zipOut, process);

                if (process.process().exitValue() != 0)
                {
                    _info.getLogger().warning("Log Downloader Plugin failed to read `${$input}` with elevated privileges.")
                }
            }
            catch (final Exception $exception)
            {
                _info.getLogger().info("Log Downloader Plugin failed to read `${$input}`: ${$exception.message}")
            }
            $zipOut.closeEntry()
        }
    }

    private void addProcessOutputToZip (final ZipOutputStream $zipOut, final ProcessHelper $process, final String $name)
    {
        final long start = System.currentTimeMillis();
        $zipOut.putNextEntry(new ZipEntry($name))
        addProcessOutputToZipEntry($zipOut, $process);
        $zipOut.closeEntry()
        
        final long elapsedMillis = System.currentTimeMillis() - start;
        if (elapsedMillis >= 10000)
        {
            _info.getLogger().warning("${$name} took ${elapsedMillis} millis to record")
        }
    }

    private void addProcessOutputToZipEntry (final ZipOutputStream $zipOut, final ProcessHelper $process)
    {
        try
        {
            $process.onInputLine({
                $zipOut << it
                $zipOut << "\n"
            })
            $process.run()
        }
        catch (final Throwable $exception)
        {
            _info.getLogger().info("Log Downloader Plugin failed to capture process output: ${$exception.message}")
        }
    }

    /*
     * Retrieves the Cattura system logs (usually under <root>/logs) timestamped
     * from $from to $to.
     */
    private Collection<FoundLogFile> getCatturaSystemLogsBetween (final Date $from, final Date $to)
    {
        return _logDir.listFiles().findAll({
            if (!it.isFile() || it.name.endsWith(".lck"))
            {
                return false
            }
            else
            {
                /*
                 * Simple check that tests last modified time
                 */
                return dateIsBetween($from, new Date(it.lastModified()), $to)
            }
        }).collect({
            return new FoundLogFile(_logDir, it)
        })
    }

    /*
     * Retrieves logs from given directory whose last modified date is between
     * $from and $to
     */
    private Collection<FoundLogFile> getLogsFromDirBetween (final File $dir, final Date $from, final Date $to, final boolean $recursive = true)
    {
        if ($dir == null || !$dir.isDirectory()) return Collections.EMPTY_LIST
        final Iterable<File> files = $recursive ? Files.fileTreeTraverser().breadthFirstTraversal($dir) : Arrays.asList($dir.listFiles())
        return files.findAll({
            return it.isFile() && dateIsBetween($from, new Date(it.lastModified()), $to)
        }).collect({
            return new FoundLogFile($dir, it)
        })
    }

    /*
     * Retrieves all capture session logs timestamped from $from to $to.
     */
    private Collection<CaptureSessionLog> getCaptureSessionLogsBetween (final Date $from, final Date $to)
    {
        return getCaptureSessionLogsInDir(_captureSessionDir) {
            return dateIsBetween($from, new Date(it.lastModified()), $to)
        }
    }

    /*
     * Retrieves all capture session logs in the given directory. $test is a closure
     * which filters out log files.
     *
     * Directory may be the root of a capture session or it may be a superdirectory
     *     containing multiple capture sessions.
     */
    private Collection<File> getCaptureSessionLogsInDir (final File $dir, final Closure $test = { return true })
    {
        final List<CaptureSessionLog> logs = []
        $dir.eachFileRecurse(FileType.FILES) {
            if ((it.name.endsWith(".log") || it.name.contains(".log.")) && !it.name.endsWith(".lck"))
            {
                /*
                 * If log file does not pass test, don't include it
                 */
                if (!$test(it))
                {
                    return
                }

                /*
                 * Collect all the available information about this log,
                 * including whether it's a log for a single node or the overall
                 * session.
                 */
                File parent = it.parentFile
                int nodeID = -1
                String pluginID = null

                while (true)
                {
                    final File graph = new File(parent, GraphBuilder.GRAPH_FILENAME)
                    if (graph.exists())
                    {
                        /*
                         * By convention, nodeID zero is the "root" node,
                         * which isn't a real processor and hence does not have
                         * an actual log
                         */
                        if (nodeID == 0)
                        {
                            break
                        }

                        logs.add(new CaptureSessionLog(
                            it, graph, parent, pluginID, nodeID
                        ))
                        break
                    }
                    else if (parent == $dir)
                    {
                        // We should have hit a graph file by now
                        // We've likely done something wrong here
                        break
                    }
                    else if (parent.name.isInteger() && nodeID == -1)
                    {
                        nodeID = parent.name as Integer
                    }
                    else if (!pluginID)
                    {
                        pluginID = parent.name
                    }

                    parent = parent.parentFile
                }
            }
        }
        return logs
    }

    private static Date argumentAsDate (final Object $arg)
    {
        if ($arg instanceof Number)
        {
            return new Date($arg.longValue())
        }
        
        return $arg
    }
    
    /*
     * Helper function to check if $between is between $before and $after.
     */
    private static boolean dateIsBetween (final Date $before, final Date $between, final Date $after)
    {
        return ($between == $before || $between.after($before)) && ($between == $after || $between.before($after))
    }

    /*
     * Translates an exception (with optional date) to a JSON-able representation
     */
    private static def exceptionToMap (final Date $date = null, final Throwable $exception)
    {
        def map = [
            type: $exception.class.name,
            message: $exception.message,
            stack: $exception.stackTrace
        ]

        if ($date)
        {
            map.timestamp = $date.time
            map.date = $date.toString()
        }

        if ($exception.cause)
        {
            map.cause = exceptionToMap($exception.cause)
        }

        return map
    }

    /*
     * Serialize the graph into our logs zip. Certain configuration
     * values which are assumed to be sensitive (publishing passwords etc)
     * are censored.
     */
    private String serializeGraphAtRootForZip (final File $sessionRoot)
    {
        final List<String> SENSITIVE_KEYS = ["key", "secret", "password", "token"]
        def graph = _json.parse(Files.toString(new File("graph.json", $sessionRoot), Charsets.UTF_8)).getOrRethrow()
        graph.nodes.each({
            def config = it.config

            if (config instanceof String)
            {
                config = _json.parse(config).getOrRethrow()
            }

            config = Maps.newLinkedHashMap(config)
            SENSITIVE_KEYS.each({
                if (config.containsKey(it))
                {
                    config.put(it, "********")
                }
            })
            it.config = _json.stringify(config).getOrRethrow()
        })
        return _json.stringify(graph).getOrRethrow()
    }

    /*
     * Value Object encapsulating a single capture session log. The object knows:
     * - Log file itself
     * - Graph file
     * - Root of the capture session
     * - Plugin ID
     * - Node ID
     *
     * These are returned from getCaptureSessionLogsInDir, which derives the above
     * information from the directory structure.
     */
    private static class CaptureSessionLog
    {
        final File _logFile
        final File _graphFile
        final File _rootFile
        final String _pluginID
        final int _nodeID

        public CaptureSessionLog (final File $logFile, final File $graphFile, final File $rootFile, final String $pluginID, final int $nodeID)
        {
            _logFile = $logFile
            _graphFile = $graphFile
            _rootFile = $rootFile
            _pluginID = $pluginID
            _nodeID = $nodeID
        }

        public File getLogFile () {return _logFile}
        public File getGraphFile () {return _graphFile}
        public File getRootFile () {return _rootFile}
        public String getPluginID () {return _pluginID}
        public int getNodeID () {return _nodeID}

        public String getZipEntryName ()
        {
            if (_logFile.getName() == "processor.log")
            {
                return nodeID > 0 ? "${_pluginID}-${_nodeID}.log" : "session.log"
            }

            return _logFile.getName()
        }

        public String toString () {return "(CaptureSessionLog ${_logFile} for node ${_nodeID} plugin ${_pluginID} at ${_rootFile})"}
    }

    private static class FoundLogFile
    {
        final File _file
        final String _entryName

        public FoundLogFile (final File $rootFile, final File $logFile)
        {
            _file = $logFile
            _entryName = $logFile.getAbsolutePath().substring($rootFile.getAbsolutePath().length() + 1)
        }

        public File getLogFile () {return _file}
        public String getEntryName () {return _entryName}
    }
}

plugin = new LogDownloaderPlugin()